use reqwest::Client;
use serde::{Deserialize, Serialize};
use tokio::io::AsyncBufReadExt;
use tokio_stream::wrappers::LinesStream;
use tokio_stream::{Stream, StreamExt};
use tokio_util::io::StreamReader;

const API_URL: &'static str = "http://localhost:11434";

pub struct OllamaAPI {
    api_url: &'static str,
    client: Client,
}

#[derive(Serialize)]
pub struct OllamaChat<'a> {
    #[serde(skip_serializing)]
    api: &'a OllamaAPI,
    model: String,
    #[serde(skip_serializing)]
    history_size: u32,
    messages: Vec<ChatMessage>,
}

#[derive(Serialize, Deserialize)]
pub enum Role {
    #[serde(rename = "system")]
    System,
    #[serde(rename = "user")]
    User,
    #[serde(rename = "assistant")]
    Assistant,
}

#[derive(Serialize, Deserialize)]
pub struct ChatMessage {
    role: Role,
    content: String,
}

#[derive(Deserialize)]
struct OllamaResponse {
    message: ChatMessage,
}

fn deserialize_chunk(chunk: Result<String, std::io::Error>) -> anyhow::Result<String> {
    match chunk {
        Err(err) => Err(anyhow::Error::new(err)),
        Ok(str) => {
            let response = serde_json::from_str::<OllamaResponse>(&str)?;
            Ok(response.message.content)
        }
    }
}

impl OllamaAPI {
    pub async fn complete<'a>(
        &self,
        chat: &OllamaChat<'a>,
    ) -> anyhow::Result<impl Stream<Item = anyhow::Result<String>>> {
        let body = serde_json::to_string(&chat)?;
        println!("Sending: {}", body);

        let request = self
            .client
            .post(String::from(API_URL) + "/api/chat")
            .body(serde_json::to_string(&chat)?)
            .build()?;
        let stream = self
            .client
            .execute(request)
            .await?
            .bytes_stream()
            .map(|x| match x {
                Err(err) => Err(std::io::Error::new(std::io::ErrorKind::Other, err)),
                Ok(bytes) => Ok(bytes),
            });
        let lines = StreamReader::new(stream).lines();
        Ok(LinesStream::new(lines).map(deserialize_chunk))
    }

    pub fn create(api_url: &'static str) -> anyhow::Result<Self> {
        let client = Client::builder().build()?;
        Ok(Self {
            api_url: api_url,
            client,
        })
    }

    pub fn create_chat(&self, model: &str) -> OllamaChat {
        OllamaChat {
            api: &self,
            history_size: 0,
            model: String::from(model),
            messages: Vec::new(),
        }
    }
}

impl<'a> OllamaChat<'a> {
    pub fn send_message(&mut self, role: Role, content: &str) {
        self.messages.push(ChatMessage {
            role,
            content: String::from(content),
        })
    }

    pub fn send_system(&mut self, content: &str) {
        self.send_message(Role::System, content);
    }

    pub fn send_user(&mut self, content: &str) {
        self.send_message(Role::User, content);
    }

    pub fn send_assistant(&mut self, content: &str) {
        self.send_message(Role::Assistant, content);
    }

    pub async fn complete(&self) -> anyhow::Result<impl Stream<Item = anyhow::Result<String>>> {
        self.api.complete(&self).await
    }
}
